/*
Copyright © 2020 Bob Callaway <bcallawa@redhat.com>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package pki

import (
	"os"
	"testing"

	"go.uber.org/goleak"
)

func TestMain(m *testing.M) {
	goleak.VerifyTestMain(m)
}

func TestFactoryNewKey(t *testing.T) {
	type TestCase struct {
		name          string
		format        string
		keyFile       string
		sigFile       string
		expectSuccess bool
	}

	testCases := []TestCase{
		{
			name:          "valid pgp",
			format:        "pgp",
			keyFile:       "pgp/testdata/valid_armored_public.pgp",
			sigFile:       "pgp/testdata/hello_world.txt.asc.sig",
			expectSuccess: true,
		},
		{
			name:          "valid minisign",
			format:        "minisign",
			keyFile:       "minisign/testdata/minisign.pub",
			sigFile:       "minisign/testdata/hello_world.txt.minisig",
			expectSuccess: true,
		},
		{
			name:          "valid x509",
			format:        "x509",
			keyFile:       "x509/testdata/ec.pub",
			sigFile:       "x509/testdata/hello_world.txt.sig",
			expectSuccess: true,
		},
		{
			name:          "valid ssh",
			format:        "ssh",
			keyFile:       "ssh/testdata/id_rsa.pub",
			sigFile:       "ssh/testdata/hello_world.txt.sig",
			expectSuccess: true,
		},
		{
			name:          "invalid ssh signature",
			format:        "ssh",
			keyFile:       "ssh/testdata/id_rsa.pub",
			sigFile:       "ssh/testdata/hello_world.txt",
			expectSuccess: false,
		},
		{
			name:          "invalid ssh key",
			format:        "ssh",
			keyFile:       "ssh/testdata/hello_world.txt",
			sigFile:       "ssh/testdata/hello_world.txt.sig",
			expectSuccess: false,
		},
		{
			format:        "bogus",
			expectSuccess: false,
		},
	}

	for _, tc := range testCases {
		tc := tc
		t.Run(tc.name, func(t *testing.T) {
			factory := NewArtifactFactory(tc.format)
			keyFile, _ := os.Open(tc.keyFile)
			_, newKeyErr := factory.NewPublicKey(keyFile)

			sigFile, _ := os.Open(tc.sigFile)
			_, newSigErr := factory.NewSignature(sigFile)

			if tc.expectSuccess {
				if newKeyErr != nil || newSigErr != nil {
					t.Errorf("unexpected error generating public key %v or signature %v", newKeyErr, newSigErr)
				}
			} else { // expect a failure{
				if newKeyErr == nil && newSigErr == nil {
					t.Error("expected error generating public key and signature. got none")
				}
			}
		})
	}
}
